import motmetrics as mm
from tqdm import tqdm
import numpy as np
import pathlib
import pandas as pd
import pickle
def drop_zero_list(row):
    return any(i != 0 for i in row)

def creat3Dpos(row):
    return [row['PosiX'], row['PosiY'], row['PosiZ']]

def MeanBubSize(row):
    return sum(row['Sizes2D'])*1000/len(row['Sizes2D']) if len(row['Sizes2D'])>0 else None

def MeanBubShape(row):
    return sum(row['Shape2D'])/len(row['Shape2D']) if len(row['Shape2D'])>0 else None

def check_visibility(row,num_cams,thresh):
    if sum([1 if row['Visibility_Cam'+str(i)]>thresh else 0 for i in range(1,num_cams+1)])>=3:
        return True
    else:
        return False

def Num_for_size(row,num_cams):
    num_for_size=0
    for i in range(1,num_cams+1):
        if row['Cam'+str(i)+'_Size']>0:
            num_for_size+=1
    return num_for_size

def TwoSphereInter(pos1:list, pos2:list, r1:float, r2:float):
    """
    https://mathworld.wolfram.com/Sphere-SphereIntersection.html
    """
    d = np.linalg.norm(np.array(pos1)-np.array(pos2))
    if r2 > r1:
        vol_total = (4. / 3. * np.pi * r1 ** 3)
    else:
        vol_total = (4. / 3. * np.pi * r2 ** 3) 
    #print(d)
    if d >= (r1 + r2):
        return None
    elif r1 > (d + r2): 
        #print('sphere 2 in sphere 1')
        vol_fraction =  (4. / 3. * np.pi * r2 ** 3) / vol_total 
    elif r2 > (d + r1):  
        #print('sphere 1 in sphere 2')
        vol_fraction =  (4. / 3. * np.pi * r1 ** 3) / vol_total
    else:
        #print('normal intersection')
        vol_fraction =  ((np.pi * (r1 + r2 - d) ** 2 * (d ** 2 + (2 * d * r1 - 3 * r1 ** 2 + 2 * d * r2 - 3 * r2 ** 2) + 6 * r1 * r2)) / (12 * d)) / vol_total
        
    return vol_fraction

def get_sorted_GT(GTDir,num_cams):
    track_files=sorted(pathlib.Path(GTDir+'/Tracks').glob('*pkl'))
    frame_vol={f:0 for f in range(len(list(pathlib.Path(GTDir+'/Cam1/').glob('*png'))))} 
    min_max={'min_x':999999,'max_x':-999999,'min_y':999999,'max_y':-999999,'min_z':999999,'max_z':-999999} 
    df_GT = pd.DataFrame()
    df_GT_sort = pd.DataFrame()
    counter=0
    for i,f in enumerate(track_files):
        with open(f, "rb") as fh:
            df=pickle.load(fh)
        df=df.reset_index()
        df['ID'] = i
        df['Frame'] = df.index
        df_GT = pd.concat([df_GT, df], ignore_index=True)
        df_test=df.copy()
        df_test['Visibility']=df_test.apply(check_visibility,axis=1,args=(num_cams,0.1))
        start=(df_test.Visibility.values==True).argmax()
        if start>0:
            df_test=df_test.iloc[start:] 
        rev_vis=df_test['Visibility'][::-1] 
        end=(rev_vis.values==True).argmax()
        if end>0:
            df_test=df_test.iloc[:len(df_test)-end] 
        last_good=0
        for frame in df_test['Frame'].unique():
            df_frame=df_test[df_test.Frame==frame]  
            inside=0
            vol=0
            for c in range(1,num_cams+1):
                if not isinstance(df_frame[f'RD_Cam{c}'].values[0],list):
                    inside+=1
                    vol+=(1/6*np.pi*df_frame[f'Cam{c}_Size'].values[0]**3)
            if inside>=3:
                #print(vol)
                frame_vol[frame]+=(vol/inside) 
                px=df_frame['PosiX'].values[0]/1000  
                py=df_frame['PosiY'].values[0]/1000  
                pz=df_frame['PosiZ'].values[0]/1000  
                if px<min_max['min_x']:
                    min_max['min_x']=px
                if px>min_max['max_x']:
                    min_max['max_x']=px

                if py<min_max['min_y']:
                    min_max['min_y']=py
                if py>min_max['max_y']:
                    min_max['max_y']=py

                if pz<min_max['min_z']:
                    min_max['min_z']=pz
                if pz>min_max['max_z']:
                    min_max['max_z']=pz
                if last_good==-1:
                    last_good=frame
            elif last_good!=-1:
                df_split=df_test[df_test.Frame<=frame-1]  
                df_split=df_split[df_split.Frame>=last_good]
                df_split['ID']=counter
                counter+=1 
                df_GT_sort = pd.concat([df_GT_sort, df_split], ignore_index=True)
                last_good=-1
        if last_good==0:
            df_test['ID']=counter 
            counter+=1 
            df_GT_sort = pd.concat([df_GT_sort, df_test], ignore_index=True)
        elif last_good!=-1: 
            df_split=df_test[df_test.Frame>=last_good]
            df_split['ID']=counter
            counter+=1 
            df_GT_sort = pd.concat([df_GT_sort, df_split], ignore_index=True)
    tot_vol=(min_max['max_x']-min_max['min_x'])*(min_max['max_y']-min_max['min_y'])*(min_max['max_z']-min_max['min_z'])
    void_frac=[]  
    for key,values in frame_vol.items():
        void_frac.append(values/tot_vol)
    
    df_tmp=df_GT_sort
    if num_cams==4:
        df_3DGT = df_tmp.loc[:,['PosiX','PosiY','PosiZ','Cam1_Size','Cam2_Size','Cam3_Size','Cam4_Size','ID','Frame']].reset_index(drop=True)
        df_3DGT['Pos3D'] = df_3DGT.apply(creat3Dpos,axis=1)
        df_3DGT['BBSize'] = ((df_3DGT.Cam1_Size + df_3DGT.Cam2_Size + df_3DGT.Cam3_Size+ df_3DGT.Cam4_Size) / 4) * 1000 ## in [mm] 
        df_3DGT = df_3DGT.drop(['PosiX','PosiY','PosiZ','Cam1_Size','Cam2_Size','Cam3_Size','Cam4_Size'],axis=1)
    else:
        df_3DGT = df_tmp.loc[:,['PosiX','PosiY','PosiZ','Cam1_Size','Cam2_Size','Cam3_Size','ID','Frame']].reset_index(drop=True)
        df_3DGT['Pos3D'] = df_3DGT.apply(creat3Dpos,axis=1)
        df_3DGT['BBSize'] = ((df_3DGT.Cam1_Size + df_3DGT.Cam2_Size + df_3DGT.Cam3_Size) / 3) * 1000 ## in [mm] 
        df_3DGT = df_3DGT.drop(['PosiX','PosiY','PosiZ','Cam1_Size','Cam2_Size','Cam3_Size'],axis=1)
    df_tmp=df_GT
    if num_cams==4:
        df_3DGT_all = df_tmp.loc[:,['PosiX','PosiY','PosiZ','Cam1_Size','Cam2_Size','Cam3_Size','Cam4_Size','ID','Frame']].reset_index(drop=True)
        df_3DGT_all['Pos3D'] = df_3DGT_all.apply(creat3Dpos,axis=1)
        df_3DGT_all['BBSize'] = ((df_3DGT_all.Cam1_Size + df_3DGT_all.Cam2_Size + df_3DGT_all.Cam3_Size+ df_3DGT_all.Cam4_Size) / 4) * 1000 ## in [mm] 
        df_3DGT_all = df_3DGT_all.drop(['PosiX','PosiY','PosiZ','Cam1_Size','Cam2_Size','Cam3_Size','Cam4_Size'],axis=1)
    else:
        df_3DGT_all = df_tmp.loc[:,['PosiX','PosiY','PosiZ','Cam1_Size','Cam2_Size','Cam3_Size','ID','Frame']].reset_index(drop=True)
        df_3DGT_all['Pos3D'] = df_3DGT_all.apply(creat3Dpos,axis=1)
        df_3DGT_all['BBSize'] = ((df_3DGT_all.Cam1_Size + df_3DGT_all.Cam2_Size + df_3DGT_all.Cam3_Size) / 3) * 1000 ## in [mm] 
        df_3DGT_all = df_3DGT_all.drop(['PosiX','PosiY','PosiZ','Cam1_Size','Cam2_Size','Cam3_Size'],axis=1)
    
    return df_3DGT,void_frac,df_3DGT_all,min_max


def prepare_output(active_tracklets,inactive_tracklets,num_cams,all_pos_raw):
    df_pred = pd.DataFrame()
    counter=0
    for Track in active_tracklets:
        Track.Positions[0]=Track.Positions[0].reshape(3)
        pos=np.stack(Track.Positions)
        BubSizes2D=Track.BubSizes2D
        BubShape2D=Track.BubShape2D
        time=Track.Ts
        Ids=Track.Ids
        if Track.pred>0 and Track.active==True:
            pos=pos[:-Track.pred] 
            BubSizes2D=BubSizes2D[:-Track.pred]
            BubShape2D=BubShape2D[:-Track.pred]  
            time=time[:-Track.pred] 
            Ids=Ids[:-Track.pred]
        df=pd.DataFrame({'PosiX':pos[:,0],'PosiY':pos[:,1],'PosiZ':pos[:,2],'Sizes2D':BubSizes2D,'Shape2D':BubShape2D,'ID':counter,'Frame':time})

        df['BBSize'] = df.apply(MeanBubSize,axis=1)
        df['BBShape'] = df.apply(MeanBubShape,axis=1)
        df['Pos3D'] = df.apply(creat3Dpos,axis=1)
        df=df.sort_values(by='Frame').interpolate()
        pos_dict={} 
        for k in range(1,num_cams+1):
            pos_dict[f'Cam{k}_YX']=[0]*len(time)
        for i,(ids,T) in enumerate(zip(Ids,time)):
            for j in range(1,num_cams+1):
                if not isinstance(ids[f'Cam{j}'],list):
                    for ii in range(len(all_pos_raw[T][f'Cam{j}'])):
                        if ids[f'Cam{j}']==all_pos_raw[T][f'Cam{j}'][ii].ID: 
                            pos_dict[f'Cam{j}_YX'][i]=all_pos_raw[T][f'Cam{j}'][ii].Position 
        for k in range(1,num_cams+1):
            df[f'Cam{k}_YX']=pos_dict[f'Cam{k}_YX']
        df_pred= pd.concat([df_pred, df], ignore_index=True)
        counter+=1

    
    for Track in inactive_tracklets:
        Track.Positions[0]=Track.Positions[0].reshape(3)
        pos=np.stack(Track.Positions)
        BubSizes2D=Track.BubSizes2D
        BubShape2D=Track.BubShape2D
        time=Track.Ts
        Ids=Track.Ids
        if Track.pred>0 and Track.active==True:
            pos=pos[:-Track.pred] 
            BubSizes2D=BubSizes2D[:-Track.pred] 
            BubShape2D=BubShape2D[:-Track.pred]
            time=time[:-Track.pred]
            Ids=Ids[:-Track.pred]
        df=pd.DataFrame({'PosiX':pos[:,0],'PosiY':pos[:,1],'PosiZ':pos[:,2],'Sizes2D':BubSizes2D,'Shape2D':BubShape2D,'ID':counter,'Frame':time})
        df['BBSize'] = df.apply(MeanBubSize,axis=1)
        df['BBShape'] = df.apply(MeanBubShape,axis=1)
        df['Pos3D'] = df.apply(creat3Dpos,axis=1)
        df=df.sort_values(by='Frame').interpolate()
        pos_dict={} 
        for k in range(1,num_cams+1):
            pos_dict[f'Cam{k}_YX']=[0]*len(time)
        for i,(ids,T) in enumerate(zip(Ids,time)):
            for j in range(1,num_cams+1):
                if not isinstance(ids[f'Cam{j}'],list):
                    for ii in range(len(all_pos_raw[T][f'Cam{j}'])):
                        if ids[f'Cam{j}']==all_pos_raw[T][f'Cam{j}'][ii].ID: 
                            pos_dict[f'Cam{j}_YX'][i]=all_pos_raw[T][f'Cam{j}'][ii].Position 
        for k in range(1,num_cams+1):
            df[f'Cam{k}_YX']=pos_dict[f'Cam{k}_YX']
        df_pred= pd.concat([df_pred, df], ignore_index=True)
        counter+=1
    
    return df_pred

def cal_metrics(df_EXP,df_GT,Num_Frame):
    acc = mm.MOTAccumulator(auto_id=True)
    for f in tqdm(range(0,Num_Frame+1,1)):
        df_exp = df_EXP[df_EXP.Frame == f]
        df_gt = df_GT[df_GT.Frame == f]    
        Id_exp = df_exp.ID.tolist()
        Id_gt = df_gt.ID.tolist()
        rows = len(Id_exp)
        cols = len(Id_gt)
        matrix = np.zeros((rows,cols),dtype=float)
        for i in range(rows):
            pos3d_exp = df_exp[df_exp.ID == Id_exp[i]].Pos3D.tolist()[0]
            r_exp = ((df_exp[df_exp.ID == Id_exp[i]].BBSize.values) / 2)[0]
            for j in range(cols):
                pos3d_gt = df_gt[df_gt.ID == Id_gt[j]].Pos3D.tolist()[0]
                r_gt = ((df_gt[df_gt.ID == Id_gt[j]].BBSize.values) / 2)[0]        
                vol_f = TwoSphereInter(pos3d_exp, pos3d_gt, r_exp, r_gt)
                if vol_f is not None:
                    vol_f=1-vol_f
                matrix[i, j] = vol_f
        acc.update(Id_exp,Id_gt,matrix)
    
    mh = mm.metrics.create()
    summary = mh.compute(acc, metrics=['num_frames', 'mota', 'motp','num_unique_objects','num_predictions','idfp','idfn','idtp','idp','idr','idf1','num_switches','num_matches','num_misses'], name='acc')  
    return summary,acc