import numpy as np
from calibration import Image_to_World,World_to_Image
from machting3D import *
from scipy.spatial.distance import cdist
import pandas as pd
from scipy.ndimage import uniform_filter1d
from utils_StarBub import RDObj

def pre_siam(Bubble,Metric,cam_num,py_calib):
    RD=Bubble.ConvertToRDC()
    RD.get_dists()
    RD.dists=uniform_filter1d(RD.dists,size=4)
    RD_2D=RD.dists*Metric[f'Cam{cam_num}'][0][0]*100
    points=np.copy(RD.points[:,:2])
    points[:,[1,0]]=points[:,[0,1]] 
    pp=Image_to_World(points.astype(float),py_calib[f'Cam{cam_num}'])
    px=pp[0,:]
    pz=pp[2,:]
    points_tran=np.zeros((64,2))
    points_tran[:,0]=pz 
    points_tran[:,1]=px
    Rdc2=RDObj(2,64,center=(np.mean(pz),np.mean(px)),points=points_tran)
    Rdc2.get_dists()
    Rdc2.dists=uniform_filter1d(Rdc2.dists,size=4)
    RD_proj=Rdc2.dists/10
    return RD_2D,RD_proj

def LMSWiener(data,order):
    dn=data[-1]
    un=data[-order-1:-1]  
    h0=np.ones(order)
    u=1/np.sum(un.conj().T*un)
    while np.abs(dn-np.dot(h0,un))>1E-8:
        h0=h0+u*un.conj().T*(dn-np.dot(h0,un))
    predict=np.dot(h0,data[-order:])
    error=-2*un*dn+2*np.dot((np.dot(un,un.conj().T)),h0.conj().T)
    return predict,np.linalg.norm(error) 

def predict_wiener(track_pos):    
    track=np.stack(track_pos)
    wienerpredx,_=LMSWiener(track[:,0],2)
    wienerpredy,_=LMSWiener(track[:,1],2)
    wienerpredz,_=LMSWiener(track[:,2],2)
    return np.array([wienerpredx,wienerpredy,wienerpredz])


def predict_and_connect_track(active_tracks,future_tracks,check_for_pred,max_dist,max_pred=2):
    future_pos=np.empty((len(future_tracks),3))
    for i in range(len(future_tracks)):  
        future_pos[i,...]=future_tracks[i].Positions[0] 
    to_remove=[] 
    for i in range(len(active_tracks)):  
        track_pos=active_tracks[i].Positions 
        track_pos[0]=track_pos[0].reshape(3)
        if active_tracks[i].pred<=max_pred and len(track_pos)>2:
            worldpos=predict_wiener(track_pos)
            worldpos=np.expand_dims(worldpos,axis=0)
            dist=cdist(worldpos,future_pos)
            if np.min(dist)<max_dist:
                idx=np.argwhere(dist==np.min(dist))[0][1]  
                to_connect=future_tracks[idx] 
                for j in range(len(check_for_pred[to_connect.last_T])):
                    if check_for_pred[to_connect.last_T][j].ID==to_connect.ID: 
                        check_for_pred[to_connect.last_T][j].add_before(active_tracks[i])
                        to_remove.append(i)

    for i in reversed(to_remove):
        del active_tracks[i] 

def predict_and_connect(tracklets,All_Bubbles,py_calib,time,max_dist,max_pred=2,px_dist=30):
    
    for i in range(len(tracklets)):  
        if tracklets[i].pred<=max_pred:
            track_pos=tracklets[i].Positions 
            track_pos[0]=track_pos[0].reshape(3)
            if len(track_pos)>2:
                worldpos=predict_wiener(track_pos)
            else:
                posx=2*track_pos[-1][0]-track_pos[-2][0] 
                posy=2*track_pos[-1][1]-track_pos[-2][1] 
                posz=2*track_pos[-1][2]-track_pos[-2][2] 
                worldpos=np.array([posx,posy,posz])
            if np.linalg.norm(worldpos-track_pos[-1])>max_dist:
                continue
            matches={} 
            all_matches={'Cam'+str(c+1):[] for c in range(len(All_Bubbles))} 
            bub_size=[] 
            bub_shape=[] 
            match_count=0
            for cam in All_Bubbles.keys():
                pos_im=World_to_Image(worldpos,py_calib[cam])
                match=[] 
                for j in range(len(All_Bubbles[cam])):
                    center1=All_Bubbles[cam][j].Position
                    dist=np.sqrt((pos_im[0]-center1[1])**2+(pos_im[1]-center1[0])**2) 
                    if dist<px_dist:
                        match.append(j)
                if len(match)==1:
                    matches[cam]=match[0] 
                    match_count+=1
            if match_count>2:
                P=np.zeros(3)
                M=np.zeros((3,3))
                for cam in matches.keys():
                    idx=matches[cam] 
                    center=All_Bubbles[cam][idx].Position
                    sight=Image_to_World(np.array([center[1],center[0]],dtype=float).reshape((1,2)),py_calib[cam])-py_calib[cam]['Tinv']
                    sight/=np.linalg.norm(sight)
                    tmp=create_tmp(sight)
                    P+=(np.dot(tmp,py_calib[cam]['Tinv'].reshape(3)))
                    M+=tmp

                worldpos2=np.dot(np.linalg.inv(M),P)
                cam=list(matches.keys())[0] 
                idx=matches[cam] 
                center=All_Bubbles[cam][idx].Position
                pos_im=World_to_Image(worldpos2.reshape(3),py_calib[cam])
                dist=np.sqrt((pos_im[0]-center[1])**2+(pos_im[1]-center[0])**2)
                if dist>100:
                    print('Yep',dist)
                    continue
                else:
                    for cam in matches.keys():
                        idx=matches[cam]
                        all_matches[cam]=All_Bubbles[cam][idx].ID
                        if isinstance(All_Bubbles[cam][idx].Diameter,float):
                            bub_size.append(All_Bubbles[cam][idx].Diameter)
                            bub_shape.append(All_Bubbles[cam][idx].Minor/All_Bubbles[cam][idx].Major)
                        else:
                            bub_size.append(All_Bubbles[cam][idx].Diameter.flatten()[0])
                            bub_shape.append(All_Bubbles[cam][idx].Minor.flatten()[0]/All_Bubbles[cam][idx].Major.flatten()[0])
                        del All_Bubbles[cam][idx]  
                    #all_matches.append(matches)
                tracklets[i].add_pos(worldpos2,time,ids=all_matches,BubSizes2D=bub_size,BubShape2D=bub_shape)
                tracklets[i].pred=0
            else:
                tracklets[i].add_prediction(worldpos,time)

                

def track(active_tracklets,All_Bubbles,py_calib,f,previous,inactive_tracklets,num_cams,max_dist,Metric,search_range=20,siamese=None,sim_thresh=0.5,max_pred=2,verbose=0,min_grad=20):
    if f>2:
        predict_and_connect(active_tracklets,All_Bubbles,py_calib,f,max_dist,max_pred=max_pred)

    triangulated_pos=[] 
    for it in range(1):    
        bubble_matches={}
        for i in range(num_cams):
            for j in range(num_cams):
                if i!=j:
                    center0on1=World_to_Image(py_calib['Cam'+str(i+1)]['Tinv'].reshape(3),py_calib['Cam'+str(j+1)])
                    bubble_matches0on1,unique_matches01,unique_counter01=match_between_view(center0on1,All_Bubbles['Cam'+str(i+1)],
                                                                                            All_Bubbles['Cam'+str(j+1)],search_range,
                                                                                            py_calib['Cam'+str(i+1)],py_calib['Cam'+str(j+1)]
                                                                                            ,min_grad=min_grad)
                    bubble_matches['Cam'+str(i+1)+'onCam'+str(j+1)]={'matches':bubble_matches0on1,'unique_matches':unique_matches01,
                                                                     'unique_counter':unique_counter01,'Siamese':{}} 
                    bubble_matches['Cam'+str(i+1)+'onCam'+str(j+1)]['unique_counter']=unique_counter01

        if num_cams<4:
            for i in range(num_cams):
                bubble_matches['Cam4onCam'+str(i+1)]={} 
                bubble_matches['Cam4onCam'+str(i+1)]['matches']={} 
                bubble_matches['Cam4onCam'+str(i+1)]['unique_matches']=[]
                bubble_matches['Cam4onCam'+str(i+1)]['unique_counter']=[]
                bubble_matches['Cam'+str(i+1)+'onCam4']={} 
                bubble_matches['Cam'+str(i+1)+'onCam4']['matches']={} 
                bubble_matches['Cam'+str(i+1)+'onCam4']['unique_matches']=[]
                bubble_matches['Cam'+str(i+1)+'onCam4']['unique_counter']=[]
        #print(bubble_matches)
        if siamese is not None:
            for i in range(1,num_cams+1):
                for j in range(1,num_cams+1):
                    if i!=j:
                        #bubble_matches[f'Cam{i}onCam{j}']['Siamese']={} 
                        for id in bubble_matches[f'Cam{i}onCam{j}']['matches']:
                            if bubble_matches[f'Cam{i}onCam{j}']['matches'][id]['unique_match']==False and len(bubble_matches[f'Cam{i}onCam{j}']['matches'][id]['matches'])>1:
                                
                                a,b=pre_siam(All_Bubbles[f'Cam{i}'][id],Metric,i,py_calib)
                                c=np.hstack((a.reshape((64,1)),b.reshape((64,1))))
                                x=[] 
                                x1=[] 
                                for match in bubble_matches[f'Cam{i}onCam{j}']['matches'][id]['matches']: 
                                    x.append(c)
                                    a1,b1=pre_siam(All_Bubbles[f'Cam{j}'][match],Metric,j,py_calib)
                                    c1=np.hstack((a1.reshape((64,1)),b1.reshape((64,1))))
                                    x1.append(c1)
                                similarity=siamese.predict([np.array(x),np.array(x1)]).flatten()
                                bubble_matches[f'Cam{i}onCam{j}']['Siamese'][id]=similarity
                                if np.min(similarity)<sim_thresh:
                                    matches=np.argmin(similarity)
                                    id2=bubble_matches[f'Cam{i}onCam{j}']['matches'][id]['matches'][matches]

                                    bubble_matches[f'Cam{i}onCam{j}']['unique_matches'].append(id)
                                    bubble_matches[f'Cam{i}onCam{j}']['unique_counter'].append(id2)

        
        all_unique=find_unique_forth_back(bubble_matches,4)
        #print(all_unique)
        if num_cams<4:
            del all_unique['Cam4'] 
        candidates=get_candidates(all_unique,bubble_matches,num_cams)
        #Check for duplicates due to Siamese
        if siamese is not None:
            to_remove={}  
            for cam in candidates.keys():
                for id in candidates[cam]:
                    for cam2 in candidates.keys():
                        if cam2!=cam:
                            for id2 in candidates[cam2]: 
                                if cam in candidates[cam2][id2]:
                                    if candidates[cam2][id2][cam]==id:                                   
                                        score1=0
                                        score2=0
                                        for cam_test in candidates[cam][id].keys():
                                            if bubble_matches[cam+'on'+cam_test]['matches'][id]['unique_match']==False: 
                                                idx=bubble_matches[cam+'on'+cam_test]['matches'][id]['matches'].index(candidates[cam][id][cam_test])
                                                score1+=bubble_matches[cam+'on'+cam_test]['Siamese'][id][idx]
                                        for cam_test in candidates[cam2][id2].keys():
                                            if bubble_matches[cam2+'on'+cam_test]['matches'][id2]['unique_match']==False: 
                                                idx=bubble_matches[cam2+'on'+cam_test]['matches'][id2]['matches'].index(candidates[cam2][id2][cam_test]) 
                                                score2+=bubble_matches[cam2+'on'+cam_test]['Siamese'][id2][idx]
                                        if score1<score2:
                                            to_remove[cam2]=id2
                                        else:
                                            to_remove[cam]=id
 
            if to_remove:
                #print(candidates)
                for key,value in to_remove.items():
                    del candidates[key][value]  
                #print('---')
                #print(candidates)
        trianged,bub_props,trian_ids=triangulate(candidates,All_Bubbles,py_calib,num_cams)
        if it>0:
            print(f'Trainged {len(trianged)} at it {it} with before {len(triangulated_pos)}')
        triangulated_pos+=trianged
        
    triangulated_pos=np.array(triangulated_pos)
    #print(f'Triangulated {len(triangulated_pos)} and active tracks {len(active_tracklets)}')
    len_triang=len(triangulated_pos)
    if len(triangulated_pos)>0:
        if len(previous['Pos'])>0:
            if f == 1:
                dist=cdist(triangulated_pos,previous['Pos'])
                dist_mask=dist<max_dist
                location=np.argwhere(dist_mask) 
                unique,index,counter=np.unique(location[:,1],return_counts=True,return_index=True)
                if np.count_nonzero(counter>1)>0:
                    dist_mask[index,unique]=counter==1 
                    print(f'Here {f}')
                to_remove=np.zeros(len(triangulated_pos),dtype=bool)
                for i in range(len(dist)):
                    location=np.argwhere(dist_mask[i])
                    if len(location)==1:
                        ids2={f'Cam{c}':trian_ids[f'Cam{c}'][i] if len(trian_ids[f'Cam{c}'])>i else [] for c in range(1,num_cams+1)}  
                        ids1={f'Cam{c}':previous['Ids'][f'Cam{c}'][int(location[0])] if len(previous['Ids'][f'Cam{c}'])>location[0] else [] for c in range(1,num_cams+1)}  
                        BubSizes=[] 
                        BubSizes2=[bub_props['size'][f'Cam{c}'][i] if len(trian_ids[f'Cam{c}'])>i else -1 for c in range(1,num_cams+1)] 
                        BubSizes2=list(filter(lambda a: a != -1, BubSizes2))
                        BubSizes.append(BubSizes2)
                        BubShape=[] 
                        BubShape2=[bub_props['aspect'][f'Cam{c}'][i] if len(trian_ids[f'Cam{c}'])>i else -1 for c in range(1,num_cams+1)] 
                        BubShape2=list(filter(lambda a: a != -1, BubShape2))
                        BubShape.append(BubShape2)
                        BubSizes1=[previous['Size'][f'Cam{c}'][int(location[0])] if len(previous['Size'][f'Cam{c}'])>location[0] else -1 for c in range(1,num_cams+1)]
                        BubSizes1=list(filter(lambda a: a != -1, BubSizes1))
                        BubSizes.append(BubSizes1)
                        BubShape1=[previous['Shape'][f'Cam{c}'][int(location[0])] if len(previous['Shape'][f'Cam{c}'])>location[0] else -1 for c in range(1,num_cams+1)]
                        BubShape1=list(filter(lambda a: a != -1, BubShape1))
                        BubShape.append(BubShape1)
                        active_tracklets.append(Tracklet(previous['Pos'][location[0]][0],triangulated_pos[i],f-1,f,BubSizes2D=BubSizes,ids1=ids1,ids2=ids2,BubShape2D=BubShape))
                        to_remove[i]=True
                triangulated_pos=np.delete(triangulated_pos,to_remove,axis=0)
                for c in range(1,4):
                    trian_ids[f'Cam{c}']=np.delete(trian_ids[f'Cam{c}'],to_remove)
                    bub_props['size'][f'Cam{c}']=np.delete(bub_props['size'][f'Cam{c}'],to_remove)
                    bub_props['aspect'][f'Cam{c}']=np.delete(bub_props['aspect'][f'Cam{c}'],to_remove)
                
                if num_cams==4:
                    trian_ids['Cam4']=np.delete(trian_ids['Cam4'],to_remove)    
                    bub_props['size']['Cam4']=np.delete(bub_props['size']['Cam4'],to_remove)  
                    bub_props['aspect']['Cam4']=np.delete(bub_props['aspect']['Cam4'],to_remove)    
            elif f>1:
            
                last_track_pos=[]
                to_mask=[] 
                for l,t in enumerate(active_tracklets):
                    last_track_pos.append(t.get_last())
                    if t.last_T==f:
                        to_mask.append(l)
                if len(last_track_pos)>0:
                    dist=cdist(triangulated_pos,np.array(last_track_pos))
                    dist_mask=dist<max_dist
                    dist_mask[:,to_mask]=False 
                    location=np.argwhere(dist_mask) 
                    unique,index,counter=np.unique(location[:,1],return_counts=True,return_index=True)
                    if np.count_nonzero(counter>1)>0:
                        dist_mask[index,unique]=counter==1 
                    to_remove=np.zeros(len(triangulated_pos),dtype=bool) 
                    for i in range(len(dist)):
                        location=np.argwhere(dist_mask[i])
                        if len(location)==1:
                            #TODO: Are we missing bub size /shape here?
                            ids={f'Cam{c}':trian_ids[f'Cam{c}'][i] if len(trian_ids[f'Cam{c}'])>i else [] for c in range(1,num_cams+1)}                          
                            active_tracklets[location[0][0]].add_pos(triangulated_pos[i],f,ids=ids)
                            active_tracklets[location[0][0]].pred=0
                            to_remove[i]=True 
                    #print(f'Before {len(triangul_pos)}')
                    triangulated_pos=np.delete(triangulated_pos,to_remove,axis=0)
                    for c in range(1,4):
                        trian_ids[f'Cam{c}']=np.delete(trian_ids[f'Cam{c}'],to_remove)
                        bub_props['size'][f'Cam{c}']=np.delete(bub_props['size'][f'Cam{c}'],to_remove)
                        bub_props['aspect'][f'Cam{c}']=np.delete(bub_props['aspect'][f'Cam{c}'],to_remove)

                    if num_cams==4:
                        trian_ids['Cam4']=np.delete(trian_ids['Cam4'],to_remove)    
                        bub_props['size']['Cam4']=np.delete(bub_props['size']['Cam4'],to_remove)  
                        bub_props['aspect']['Cam4']=np.delete(bub_props['aspect']['Cam4'],to_remove)   
                        
                dist=cdist(triangulated_pos,previous['Pos'])
                dist_mask=dist<max_dist
                location=np.argwhere(dist_mask) 
                unique,index,counter=np.unique(location[:,1],return_counts=True,return_index=True)
                
                if np.count_nonzero(counter>1)>0:
                    dist_mask[index,unique]=counter==1 
                to_remove=np.zeros(len(triangulated_pos),dtype=bool) 
                for i in range(len(dist)):
                    location=np.argwhere(dist_mask[i])
                    if len(location)==1:
                        ids2={f'Cam{c}':trian_ids[f'Cam{c}'][i] if len(trian_ids[f'Cam{c}'])>i else [] for c in range(1,num_cams+1)}  
                        ids1={f'Cam{c}':previous['Ids'][f'Cam{c}'][int(location[0])] if len(previous['Ids'][f'Cam{c}'])>location[0] else [] for c in range(1,num_cams+1)}  
                        BubSizes=[] 
                        BubSizes2=[bub_props['size'][f'Cam{c}'][i] if len(trian_ids[f'Cam{c}'])>i else -1 for c in range(1,num_cams+1)] 
                        BubSizes2=list(filter(lambda a: a != -1, BubSizes2))
                        BubSizes.append(BubSizes2)
                        BubShape=[] 
                        BubShape2=[bub_props['aspect'][f'Cam{c}'][i] if len(trian_ids[f'Cam{c}'])>i else -1 for c in range(1,num_cams+1)] 
                        BubShape2=list(filter(lambda a: a != -1, BubShape2))
                        BubShape.append(BubShape2)
                        BubSizes1=[previous['Size'][f'Cam{c}'][int(location[0])] if len(previous['Size'][f'Cam{c}'])>location[0] else -1 for c in range(1,num_cams+1)]
                        BubSizes1=list(filter(lambda a: a != -1, BubSizes1))
                        BubSizes.append(BubSizes1)
                        BubShape1=[previous['Shape'][f'Cam{c}'][int(location[0])] if len(previous['Shape'][f'Cam{c}'])>location[0] else -1 for c in range(1,num_cams+1)]
                        BubShape1=list(filter(lambda a: a != -1, BubShape1))
                        BubShape.append(BubShape1)
                        active_tracklets.append(Tracklet(previous['Pos'][location[0]][0],triangulated_pos[i],f-1,f,BubSizes2D=BubSizes,ids1=ids1,ids2=ids2,BubShape2D=BubShape))
                        to_remove[i]=True 
                triangulated_pos=np.delete(triangulated_pos,to_remove,axis=0)
                for c in range(1,4):
                    trian_ids[f'Cam{c}']=np.delete(trian_ids[f'Cam{c}'],to_remove)
                    bub_props['size'][f'Cam{c}']=np.delete(bub_props['size'][f'Cam{c}'],to_remove)
                    bub_props['aspect'][f'Cam{c}']=np.delete(bub_props['aspect'][f'Cam{c}'],to_remove)
                
                if num_cams==4:
                    trian_ids['Cam4']=np.delete(trian_ids['Cam4'],to_remove)    
                    bub_props['size']['Cam4']=np.delete(bub_props['size']['Cam4'],to_remove)  
                    bub_props['aspect']['Cam4']=np.delete(bub_props['aspect']['Cam4'],to_remove)  

                for t in active_tracklets:
                    t.check_active(f,max_pred=max_pred)
                    if t.active==False:
                        inactive_tracklets.append(t)
                        active_tracklets.remove(t)
    if verbose>0:
        print(f'Triangulated {len_triang}, not tracked {len(triangulated_pos)} and active tracks {len(active_tracklets)}')
    previous={} 
    previous['Pos']=triangulated_pos
    previous['Ids']=trian_ids
    previous['Size']=bub_props['size'] 
    previous['Shape']=bub_props['aspect'] 
    return active_tracklets,previous,inactive_tracklets

def write_tracklets(tracklets,counter,trackdir,num_cams):
    for i in range(len(tracklets)):
        if len(tracklets[i].Ts)>2:
            cam1,cam2,cam3,cam4,pX,pY,pZ=[],[],[],[],[],[],[] 
            for cams,posi in zip(tracklets[i].Ids,tracklets[i].Positions):
                cam1.append(cams['Cam1'])
                cam2.append(cams['Cam2'])
                cam3.append(cams['Cam3'])
                if num_cams==4:
                    cam4.append(cams['Cam4'])
                pX.append(posi[0])
                pY.append(posi[1])
                pZ.append(posi[2])
            df=pd.DataFrame({'Timestep':tracklets[i].Ts,'Cam1':cam1,'Cam2':cam2,'Cam3':cam3,
                             'PosiX':pX,'PosiY':pY,'PosiZ':pZ})
            if num_cams==4:
                df['Cam4']=cam4
            df.to_pickle(trackdir+'Track_'+str(counter)+'.pkl')
            counter+=1

def reverse_tracks(active_tracklets,inactive_tracklets,num_frames):
    check_for_pred={i:[] for i in range(num_frames+1)}
    check_for_connect={i:[] for i in range(num_frames+1)}

    for track in active_tracklets:
        #if track.Ts[0]>0: 
        pos=track.Positions
        pos[0]=pos[0].reshape(3)
        BubSizes2D=track.BubSizes2D
        BubShape2D=track.BubShape2D
        time=track.Ts
        Ids=track.Ids
        if track.pred>0 and track.active==True:
            pos=pos[:-track.pred] 
            BubSizes2D=BubSizes2D[:-track.pred] 
            BubShape2D=BubShape2D[:-track.pred]
            time=time[:-track.pred] 
            Ids=Ids[:-track.pred]
        BubSizes2D=BubSizes2D[::-1]
        BubShape2D=BubShape2D[::-1]
        Ids=Ids[::-1]
        Ts=num_frames-np.array(time)
        Ts=Ts[::-1].tolist()
        t=Tracklet(pos[-1],pos[-2],Ts[0],Ts[1],BubSizes2D[:2],Ids[0],Ids[1],BubShape2D[:2]) 
        t.add_list(pos[::-1][2:],Ts[len(Ts)-1],Ts[2:],BubSizes2D[2:],Ids[2:],BubShape2D[2:])
        check_for_pred[t.last_T].append(t) 
    for i,track in enumerate(inactive_tracklets):
        #if track.Ts[0]>0: 
        pos=track.Positions
        pos[0]=pos[0].reshape(3)
        BubSizes2D=track.BubSizes2D
        BubShape2D=track.BubShape2D
        time=track.Ts
        Ids=track.Ids
        if track.pred>0 and track.active==True:
            pos=pos[:-track.pred] 
            BubSizes2D=BubSizes2D[:-track.pred] 
            BubShape2D=BubShape2D[:-track.pred]
            time=time[:-track.pred] 
            Ids=Ids[:-track.pred]
        BubSizes2D=BubSizes2D[::-1]
        BubShape2D=BubShape2D[::-1]
        Ids=Ids[::-1]
        Ts=num_frames-np.array(time)
        Ts=Ts[::-1].tolist()
        t=Tracklet(pos[-1],pos[-2],Ts[0],Ts[1],BubSizes2D[:2],Ids[0],Ids[1],BubShape2D[:2]) 
        t.add_list(pos[::-1][2:],Ts[len(Ts)-1],Ts[2:],BubSizes2D[2:],Ids[2:],BubShape2D[2:])
        t.ID=i
        check_for_pred[t.last_T].append(t)     
        check_for_connect[t.Ts[0]].append(t) 
    return check_for_pred,check_for_connect

class Tracklet():
    def __init__(self,PosT1,PosT2,T1,T2,BubSizes2D=[],ids1=[],ids2=[],BubShape2D=[]):
        self.Positions=[PosT1,PosT2] 
        self.active=True
        self.last_T=T2
        self.Ts=[T1,T2] 
        self.Ids=[ids1,ids2]
        self.pred=0 
        self.BubSizes2D=BubSizes2D
        self.BubShape2D=BubShape2D
        self.color=tuple((np.random.choice(range(255),size=3))/255.0)
        self.ID=-1

    def get_last(self):
        return self.Positions[-1]

    def add_pos(self,Pos,T2,BubSizes2D=[],ids=None,BubShape2D=[]):
        self.Positions.append(Pos)
        self.last_T=T2
        self.Ts.append(T2)
        self.Ids.append(ids)
        self.BubSizes2D.append(BubSizes2D)
        self.BubShape2D.append(BubShape2D)
        
    def add_list(self,Pos,T2,Ts,BubSizes2D=[],ids=None,BubShape2D=[]):
        self.Positions.extend(Pos)
        self.last_T=T2
        self.Ts.extend(Ts)
        self.Ids.extend(ids)
        self.BubSizes2D.extend(BubSizes2D)
        self.BubShape2D.extend(BubShape2D)

    def check_active(self,T,max_pred=2):
        if self.last_T!=T or self.pred>=max_pred:
            self.active=False
            if self.pred>=max_pred:
                self.Positions=self.Positions[:-self.pred] 
                self.Ts=self.Ts[:-self.pred] 
                self.Ids=self.Ids[:-self.pred] 
                self.BubSizes2D=self.BubSizes2D[:-self.pred] 
                self.BubShape2D=self.BubShape2D[:-self.pred] 

    def add_prediction(self,Pos,T2):
        self.Positions.append(Pos)
        self.last_T=T2
        self.Ts.append(T2)
        self.pred+=1
        self.Ids.append({k:[] for k in self.Ids[0].keys()})
        self.BubSizes2D.append([])
        self.BubShape2D.append([])

    def add_before(self,track):
        self.Positions=track.Positions+self.Positions
        self.Ts=track.Ts+self.Ts
        self.Ids=track.Ids+self.Ids
        self.BubSizes2D=track.BubSizes2D+self.BubSizes2D
        self.BubShape2D=track.BubShape2D+self.BubShape2D
